#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : ccn+psf+psr.py
# Contents: Pattern Spectrum Filtering and Pattern Set Reduction
#           (for event sequences/parallel point processes/CoCoNAD)
# Author  : Christian Borgelt
# History : 2013.11.22 file created
#           2013.11.23 first version completed
#           2013.12.03 adapted to new function genpsp()
#           2013.12.05 some functions exported to patspec.py, patred.py
#           2013.12.06 filtering changed to detection border
#           2014.01.29 file reading made more general (separators etc.)
#           2014.03.06 estimation of a pattern spectrum added
#           2014.05.09 surrogate data via block permutations added
#           2014.05.12 pattern spectrum border passed to coconad
#           2014.07.16 targets for item sequences instead of sets added
#           2014.11.29 added option for no pattern spectrum filtering
#           2015.08.12 adapted to modified pattern set reduction
#           2016.10.05 target pattern types and subtypes added
#-----------------------------------------------------------------------
from sys     import argv, stderr, path, version_info
from os.path import join
from time    import time
from math    import floor, ceil
try:
    from coco    import coconad, genpsp, estpsp, psp2bdr, patred
except ImportError:
    COCODIR = join('..', 'coconad')
    if COCODIR not in path: path.insert(0, COCODIR)
    from coconad import coconad
    FIMDIR = join('..', 'fim')
    if FIMDIR  not in path: path.insert(0, FIMDIR)
    from patspec import genpsp, psp2bdr
    from estpsp  import estpsp
    from patred  import patred
    print('Attention: Could not import from PyCoCo library!')
    print('           Using Python replacement (very slow)!')

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# File Reading Function
#-----------------------------------------------------------------------

def getrec (f, recseps, fldseps, blanks, comment):
    '''Read a record from a file and split it into fields.
f       file to read from
recseps record  separators
fldseps field   separators
blanks  blank   characters
comment comment characters
returns a list of the fields read or None at end of file'''
    c = f.read(1).decode()      # read the next character
    while c != '' and c in comment:
        while c != '' and c not in recseps:
            c = f.read(1).decode() # skip comment until record separator
        c = f.read(1).decode()  # consume the record separator
    if c == '': return None     # if at end of file, abort
    rec = []                    # initialize the record
    sep = fldseps +recseps      # combine all separators
    while 1:                    # field read loop
        while c != '' and c in blanks:
            c = f.read(1).decode() # skip leading blanks
        fld = ''                # initialize the field
        while c != '' and c not in sep:
            fld += c            # append character to the field
            c = f.read(1).decode()  # and read the next character
        fld = fld.strip(blanks) # remove leading/trailing blanks */
        if len(fld) > 0:        # if the field is not empty,
            rec.append(fld)     # append the field to the record
        if c == '' or c in recseps: break
        c = f.read(1).decode()  # consume field separator
    return rec                  # return the record

#-----------------------------------------------------------------------
# Main Program
#-----------------------------------------------------------------------

def error (msg):
    stderr.write(msg); exit()   # print an error message and abort

#-----------------------------------------------------------------------

if __name__ == '__main__':
    desc    = 'Pattern Spectrum Filtering and Pattern Set Reduction\n' \
            + '(for event sequences/parallel point processes)'
    version = 'version 1.11 (2016.10.05)        ' \
            + '(c) 2013-2016   Christian Borgelt'
    args    = []                # list of program arguments/options
    opts    = {'T': ['target', 'i' ],  # target pattern type
               't': ['subori', 'c' ],  # target subtype for original
               'u': ['subsur', 'f' ],  # target subtype for surrogates
               'w': ['width', 0.003],  # width of window/region
               's': ['supp',  2.0  ],  # minimum support (absolute)
               'm': ['zmin',  2    ],  # minimum number of items
               'n': ['zmax', -1    ],  # maximum number of items
               'x': ['pex',   True ],  # perfect extension flag
               'a': ['beg',  -oo   ],  # minimum of point/time range
               'z': ['end',   oo   ],  # minimum of point/time range
               'g': ['surr',  'e'  ],  # surrogate generation method
               'c': ['cnt',   1000 ],  # number of surrogate data sets
               'd': ['rand',  'r'  ],  # random function for surrogates
               'j': ['sigma', 0.005],  # sigma for random function
               'B': ['delta', 0.03 ],  # block size (time interval)
               'e': ['alpha', 0.5  ],  # probability dispersion factor
               'i': ['smpls', 1000 ],  # samples per item set size
               'S': ['seed',  0    ],  # seed for random numbers
               'Z': ['cpus',  0    ],  # number of cores/cpus to use
               'R': ['pred',  'S'  ],  # pattern set reduction
               'q': ['pssep', ' '  ],  # pattern spectrum separator
               'h': ['ishdr', ''   ],  # record header  for output
               'k': ['isep',  ' '  ],  # item separator for output
               'v': ['outfmt',' (%d)'],# output format for support
               'l': ['train', False],  # one train per line/record
               'y': ['item',  True ],  # item in line, item first
               'r': ['recseps', '\\n' ],     # record  separators
               'f': ['fldseps', ' ,\\t' ],   # field   separators
               'b': ['blanks',  ' \\t\\r' ], # blank   characters
               'C': ['comment', '#' ],       # comment characters
               'P': ['pspfn',  ''  ] } # name of pattern spectrum file

    if len(argv) <= 1:          # if no arguments are given
        opts = dict([(o,opts[o][1]) for o in opts])
        print('usage: ccn+psf+psr.py [options] infile [outfile]')
        print(desc); print(version)
        print('-T#      target pattern type                    '
                      +'(default: %s)' % opts['T'])
        print('         i    item sets (item order is ignored)')
        print('         p    partial permutations '
                           +'(without repetition)')
        print('         s    sequences '
                           +'(with & without repetition)')
        print('-t#      target subtype for original data       '
                      +'(default: %s)' % opts['t'])
        print('-u#      target subtype for surrogate data      '
                      +'(default: %s)' % opts['u'])
        print('         f/a  all      frequent  patterns')
        print('         c    closed  (frequent) patterns')
        print('         m    maximal (frequent) patterns')
        print('-w#      width of time window/maximum distance  '
                      +'(default: %g)' % opts['w'])
        print('-s#      minimum support of an item set         '
                      +'(default: %g)' % opts['s'])
        print('-m#      minimum number of items per item set   '
                      +'(default: %d)' % opts['m'])
        print('-n#      maximum number of items per item set   '
                      +'(default: no limit)')
        print('-x       do not prune with perfect extensions   '
                      +'(default: prune)');
        print('-a#      start of point/time range              '
                      +'(default: auto)')
        print('-z#      end   of point/time range              '
                      +'(default: auto)')
        print('-g#      surrogate generation method            '
                      +'(default: %s)' % opts['g'])
        print('         x    none (no pattern spectrum filtering)')
        print('         f    none (read pattern spectrum from file)')
        print('         i    identity (keep original data)')
        print('         r    point/spike time randomization')
        print('         d    point/spike dithering/displacement')
        print('         s    train shifting/dithering')
        print('         k    sampling from a kernel estimate')
        print('         p    dithered item permutation')
        print('         b    dithered blocked item permutations')
        print('         e    estimate pattern spectrum (no surrogates)')
        print('-c#      number of surrogate data sets          '
                      +'(default: %d)' % opts['c'])
        print('         (if <= 0, the pattern spectrum is read '
                      +'from a file)')
        print('-d#      random function type for displacements '
                      +'(default: %s)' % opts['d'])
        print('         u/r  uniform/rectangular density')
        print('         t    symmetric triangular density')
        print('         g/n  Gaussian/normal density')
        print('-j#      parameter for random function density  '
                      +'(default: %g)' % opts['j'])
        print('         (standard deviation or half base width '
                      +'of rectangle/triangle)')
        print('-B#      block size for block permutations      '
                      +'(default: %g)' % opts['B'])
        print('         (only used with option -gb)')
        print('-Z#      number of cpus/processor cores to use  '
                      +'(default: %d)' % opts['Z'])
        print('         (a value <= 0 means all cpus reported '
                      +'as available)')
        print('-e#      probability dispersion factor          '
                      +'(default: %g)' % opts['e'])
        print('         (only used with option -ge)')
        print('-i#      number of item set samples per size    '
                      +'(default: %d)' % opts['i'])
        print('         (only used with option -ge)')
        print('-R#      pattern set reduction method           '
                      +'(default: %s)' % opts['R'])
        print('         x     none (keep all patterns after filtering)')
        print('         c     excess coincidences    '
                      +'(zb,cb-ca)')
        print('         C     excess coincidences    '
                      +'(zb,cb-ca+1)')
        print('         i     excess items/neurons   '
                      +'(za-zb+2,ca)')
        print('         s     covered points/spikes  '
                      +'za*ca : zb*cb')
        print('         S     covered points/spikes  '
                      +'(za-1)*ca : (zb-1)*cb')
        print('         l     combined lenient       '
                      +'(C+i+s, break only rejection tie)')
        print('         L     combined lenient       '
                      +'(C+i+S, break only rejection tie)')
        print('         t     combined strict        '
                      +'(C+i+s, always force decision)')
        print('         T     combined strict        '
                      +'(C+i+S, always force decision)')
        print('-S#      seed for pseudo-random numbers         '
                      +'(default: %d)' % opts['S'])
        print('-P#      name of pattern spectrum file          '
                      +'(default: none)')
        print('-q#      column separator for pattern spectrum  '
                      +'(default: "%s")' % opts['q'])
        print('-h#      record  header for output              '
                      +'(default: "%s")' % opts['h'])
        print('-k#      item separator for output              '
                      +'(default: "%s")' % opts['k'])
        print('-v#      output format for pattern support      '
                      +'(default: "%s")' % opts['v'])
        print('-l       one train per record (item and points) '
                      +'(default: id/time pairs)')
        print('-y       no item (with -l) or item after point  '
                      +'(default: item first)')
        print('-r#      record  separators                     '
                      +'(default: "%s")' % opts['r'])
        print('-f#      field   separators                     '
                      +'(default: "%s")' % opts['f'])
        print('-b#      blank   characters                     '
                      +'(default: "%s")' % opts['b'])
        print('-C#      comment characters                     '
                      +'(default: "%s")' % opts['C'])
        print('infile   file to read trains from               '
                      +'[required]')
        print('outfile  file to write found item sets to       '
                      +'[optional]')
        exit()                  # print usage message and abort

    stderr.write('ccn+psf+psr.py') # print a startup message
    stderr.write(' - ' +desc +'\n' +version +'\n')

    # --- evaluate program options and arguments ---
    for a in argv[1:]:          # traverse the program arguments
        if a[0] != '-':         # check for a fixed argument
            if len(args) < 2: args.append(a); continue
            else: error('too many arguments\n')
        if a[1] not in opts:    # check for an option
            error('unknown option: %s\n' % a[:2])
        o = opts[a[1]]          # get the corresponding option
        v = a[2:]               # and get the option argument
        if   isinstance(o[1],True.__class__): o[1] = not o[1]
        elif isinstance(o[1], (0).__class__): o[1] = int(v)
        elif isinstance(o[1], 0.0.__class__): o[1] = float(v)
        else:                                 o[1] = v
    if len(args) < 1: error('missing arguments\n')
    opts = dict([opts[o] for o in opts])

    target  = opts['target']    # target pattern type for CoCoNAD
    subori  = opts['subori']    # target subtype for original
    subsur  = opts['subsur']    # target subtype for surrogates
    width   = opts['width']     # time window width/maximum distance
    supp    = opts['supp']      # minimum support for CoCoNAD
    zmin    = opts['zmin']      # minimum size of item sets
    zmax    = opts['zmax']      # maximum size of item sets
    pex     = opts['pex']       # perfect extension flag
    beg     = opts['beg']       # beginning of allowed span of spikes
    end     = opts['end']       # end       of allowed span of spikes
    surr    = opts['surr']      # surrogate data generation method
    rand    = opts['rand']      # random displacement function
    sigma   = opts['sigma']     # random function parameter
    delta   = opts['delta']     # block size for blocked permutations
    alpha   = opts['alpha']     # probability dispersion factor
    smpls   = opts['smpls']     # number of item set samples per size
    seed    = opts['seed']      # random seed (0: use time)
    cnt     = opts['cnt']       # number of surrogate data sets
    cpus    = opts['cpus']      # number of cpus/cores to use
    pred    = opts['pred']      # pattern set reduction method
    pssep   = opts['pssep']     # pattern spectrum separator
    ishdr   = opts['ishdr']     # record header  for output
    isep    = opts['isep']      # item separator for output
    outfmt  = opts['outfmt']    # output format for support
    tpr     = opts['train']     # flag for one train per line
    item    = opts['item']      # and flag for item first
    recseps = opts['recseps']   # record  separators
    fldseps = opts['fldseps']   # field   separators
    blanks  = opts['blanks']    # blank   characters
    comment = opts['comment']   # comment characters
    pspfn   = opts['pspfn']     # pattern spectrum file name
    if supp  <  0: supp = -supp  # check the parameters
    if zmin  <= 0: error('invalid minimum size %d\n'    % zmin)
    if width <  0: error('invalid window width %g\n'    % width)
    if sigma <  0: error('invalid sigma parameter %g\n' % sigma)
    if delta <= 0: error('invalid block size %g\n'      % delta)
    if surr not in 'xfirdskpbe':# check surrogate generation method
        error('invalid surrogate data generation %s\n' % surr)
    if rand not in 'urtgn':     # check random displacement function
        error('invalid random function type %s\n'      % rand)
    if pred not in 'xcCisSlLtT':# check pattern set reduction
        error('invalid pattern set reduction %s\n'     % pred)
    if surr == 'x' or surr == 'f': cnt = 0
    x = [recseps,fldseps,blanks,comment]
    if version_info[0] >= 3:    # decode ASCII escape sequences
        x = [bytes(s, 'utf-8').decode('unicode_escape') for s in x]
    else:                       # decode ASCII escape sequences
        x = [s.decode('string_escape') for  s in x]
    recseps,fldseps,blanks,comment = x

    # --- evaluate target type and subtypes --- */
    if   target in ['i','s','set','sets','itemset','itemsets']:
        target = 'i'
    elif target in ['p','o','perm','perms','permut','permuts']:
        target = 'p'
    elif target in ['s','q','seq','seqs','sequence','sequences']:
        target = 's'
    else: error('invalid target type %s\n' % target)
    if   subori in ['f','s','a','all','frq','freq']: subori = 'f'
    elif subori in ['c','cls','clsd','closed']:      subori = 'c'
    elif subori in ['m','max','maxi','maximal']:     subori = 'm'
    else: error('invalid target subtype %s\n' % subori)
    trgori = subori +target     # target for original data
    if   subsur in ['f','s','a','all','frq','freq']: subsur = 'f'
    elif subsur in ['c','cls','clsd','closed']:      subsur = 'c'
    elif subsur in ['m','max','maxi','maximal']:     subsur = 'm'
    else: error('invalid target subtype %s\n' % subsur)
    trgsur = subsur +target     # target for surrogate data

    # --- read the data set to analyze ---
    t = time()                  # start timer, print log message
    stderr.write('reading %s ... ' % args[0])
    trains = dict(); k = 1      # initialize the train dictionary
    with open(args[0], 'rb') as inp:
        while 1:                # record read loop
            rec = getrec(inp, recseps, fldseps, blanks, comment)
            if rec == None: break
            if not rec: continue# read the next record
            if tpr:             # if one train per line/record
                if item: n = rec[0]; rec = rec[1:]
                else:    n = k+1# get the item/neuron identifier
                trains[n] = [x for x in [float(x) for x in line]
                             if x >= beg and x <= end]
                k += 1          # collect the trains
            else:               # if item/ point pairs
                if item: n,s = rec[:2]
                else:    s,n = rec[:2]
                s = float(s)    # collect points per item/neuron
                if s < beg or s > end: continue
                if n in trains: trains[n].append(s)
                else:           trains[n] = [s]
    trains = trains.items()     # convert to pairs (item, list of times)
    points = sum([len(s) for n,s in trains])
    stderr.write('[%d item(s), %d point(s)]' % (len(trains), points))
    stderr.write(' done [%.2fs].\n' % (time()-t))
    if beg <= -oo: beg = floor(min(min(t) for n,t in trains))
    if end >= +oo: end = ceil (max(max(t) for n,t in trains))

    # --- read or generate pattern spectrum ---
    t = time()                  # start timer, print log message
    if   surr == 'x':           # if no pattern spectrum filtering
        psp = dict()            # use an empty pattern spectrum
    elif surr == 'f':           # if to read a pattern spectrum
        stderr.write('reading %s ... ' % pspfn)
        psp = dict()            # initialize a pattern spectrum
        with open(pspfn, 'r') as inp:
            for line in inp:    # read pattern spectrum from file
                z,c,n = line.split()
                psp[int(z),int(c)] = float(n)
    elif surr == 'e':           # if to estimate a pattern spectrum
        stderr.write('estimating pattern spectrum ... '); stderr.flush()
        psp = estpsp (trains, trgsur, width, supp, zmin, zmax, '#',
                      cnt, alpha, smpls, seed)
    else:                       # if to generate a pattern spectrum
        stderr.write('generating pattern spectrum ... '); stderr.flush()
        psp = genpsp(trains, trgsur, width, supp, zmin, zmax, '#',
                     'r', ('' if pex else 'x'), cnt, beg, end,
                     surr, rand, sigma, delta, seed, cpus)
    stderr.write('[%d signature(s)]' % len(psp))
    stderr.write(' done [%.2fs].\n' % (time()-t))
    border = psp2bdr(psp)       # extract the decision border */

    # --- save generated pattern spectrum ---
    if cnt > 0 and pspfn != "": # if file name for pattern spectrum
        t = time()              # start timer, print log message
        stderr.write('writing %s ... ' % pspfn)
        with open(pspfn, 'w') as out: # write pattern spectrum to a file
            for s in sorted([(z,c,psp[z,c]) for z,c in psp]):
                out.write(('%d'+pssep+'%d'+pssep+'%.16g\n') % s)
        stderr.write('[%d signature(s)]' % len(psp))
        stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- analyze original data set ---
    if len(args) < 2: exit()    # check for an output file name
    t = time()                  # start timer, print log message
    stderr.write('analyzing original data ... ')
    pats = coconad(trains, trgori, width, supp, zmin, zmax, 'a',
                   ('' if pex else 'x'), border=border)
    stderr.write('[%d pattern(s)]' % len(pats))
    stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- pattern set reduction ---
    if target == 'i' and pred != 'x': # if to reduce the found patterns
        t = time()              # start timer, print log message
        stderr.write('reducing pattern set ... ')
        pats = patred(pats, pred, border, True)
        stderr.write('[%d pattern(s)]' % len(pats))
        stderr.write(' done [%.2fs].\n' % (time()-t))

    # --- write output file ---
    t = time()                  # start timer, print log message
    stderr.write('writing %s ... ' % args[1])
    with open(args[1], 'w') as out:
        for p,c in pats:        # traverse the reduced patterns
            out.write(ishdr)    # print the record header
            p = sorted(p)       # sort the items in the pattern
            for i in p[:-1]: out.write(str(i)+isep)
            out.write(p[-1])    # write the items of the pattern
            out.write(outfmt % c)
            out.write('\n')     # write the support information
    stderr.write('[%d pattern(s)]' % len(pats))
    stderr.write(' done [%.2fs].\n' % (time()-t))
